class Comparator:
    TEMPLATE = NotImplemented

    def __init__(self, value):
        self.value = value

    def calculate(self, field_name, field_type, *args, **kwargs):
        return self.TEMPLATE.format(field_name=field_name, value=self.value)


class NumericComparator(Comparator):
    OPERATOR = NotImplemented
    TEMPLATE = '{field_name} {operator}{equal} {value}'

    def __init__(self, value, equal=False):
        super().__init__(value)
        self.__equal = equal

    def calculate(self, field_name, field_type, *args, **kwargs):
        return self.TEMPLATE.format(
            field_name=field_name, operator=self.OPERATOR, equal='=' if self.__equal else '', value=self.value)


class Grater(NumericComparator):
    OPERATOR = '>'


class Lower(NumericComparator):
    OPERATOR = '<'


class Not(Comparator):
    TEMPLATE = '{field_name} {operator} {value}'

    def calculate(self, field_name, field_type, *args, **kwargs):
        if isinstance(field_type, bool):
            if self.value:
                operator = 'is'
            else:
                operator = 'not'

        else:
            operator = '!='

        return self.TEMPLATE.format(field_name=field_name, operator=operator, value=self.value)



from jija_orm.executors import templates
from jija_orm.executors import executors

table_selector = executors.Executor(
    'pg_catalog.pg_tables',
    templates.Select(name='tablename'),
    templates.Where(
        schemaname__not='pg_catalog\' and schemaname != \'information_schema',
        tablename__not='jija_orm',
        tablename__startswith=executors.Param('app')
    ),
    use_model=False
)

columns_selector = executors.Executor(
    'information_schema.columns',
    templates.Select(
        name='column_name', default='column_default', null='is_nullable',
        type='udt_name', max_length='character_maximum_length'
    ),
    templates.Where(table_name=executors.Param('table')),
    use_model=False
)

migration_id_selector = executors.Executor(
    'jija_orm',
    templates.Select(id='max(id)'),
    use_model=False
)

id_setter = executors.Executor(
    'jija_orm',
    templates.Insert(id=executors.Param('id')),
    use_model=False
)



import asyncio
import importlib
import os.path

from typing import Optional

from jija_orm import utils, models
from jija_orm.executors import executors
from jija_orm.executors import templates
from jija_orm.migrator import templates as migration_templates
from jija_orm import exceptions


class App:
    def __init__(
            self, *, name: str, path: str = None, models_modules: Optional[str] = None,
            migration_dir: Optional[str] = None):

        """
        :param name: App name
        :param path: App path by base path from core, if path is None it will be same as name
        :param models_modules: List of models files, if it is None it will be [models]
        :param migration_dir: Directory for app migrations, if it is None it will be "migrations.{app_name}"
        """

        self.__name = name

        self.__path = path or name
        self.__models_modules = self.__validate_models(models_modules)
        self.__migration_dir = self.__validate_migration_dir(migration_dir)

        self.__models = None

    @property
    def name(self):
        return self.__name

    @property
    def migrations_dir(self):
        return self.__migration_dir

    @property
    def models(self):
        return self.__models

    @staticmethod
    def __validate_models(models_modules):
        if models_modules is None:
            return ['models']

        if not isinstance(models_modules, (list, tuple)):
            raise exceptions.AppConfigTypeError(models_modules)

        return models_modules

    def __validate_migration_dir(self, migration_dir):
        if migration_dir:
            return migration_dir

        return f'migrations.{self.__name}'

    def check(self, base_path):
        for app_models in self.__models_modules:
            path = base_path.joinpath(self.__path.replace('.', '/'), app_models.replace('.', '/') + '.py')
            if not os.path.exists(path):
                raise exceptions.AppConfigImportError(path)

        self.__create_migrations_dir()

    def load(self):
        self.__load_models()
        asyncio.run(self.__load_migrations_table())

    async def async_load(self):
        self.__load_models()
        await self.__load_migrations_table()

    def __load_models(self):
        app_models = {}
        for model_module in self.__models_modules:
            module = importlib.import_module(f'{self.__path}.{model_module}')
            app_models.update(map(
                lambda model_class: (model_class.get_name(), model_class),
                utils.collect_subclasses(module, models.Model)
            ))

        for model_name in app_models:
            app_models[model_name].init_managers()

        self.__models = app_models

    @staticmethod
    async def __load_migrations_table():
        executor = executors.Executor(
            'pg_catalog.pg_tables',
            templates.Select(name='tablename'),
            templates.Where(tablename='jija_orm',
                            schemaname__not='pg_catalog\' and schemaname != \'information_schema'),
            use_model=False
        )

        if len(await executor.execute()) == 0:
            migration = migration_templates.ModelMigration(
                'jija_orm', migration_templates.Action.CREATE,
                [
                    migration_templates.FieldMigration(
                        'id', migration_templates.Action.CREATE,
                        [
                            migration_templates.AttributeMigration('pk', True),
                            migration_templates.AttributeMigration('null', False),
                            migration_templates.AttributeMigration('type', 'int8')
                        ]
                    )
                ],

            )

            await migration.execute()

    def __create_migrations_dir(self):
        current_path = []
        path = self.__migration_dir.split('.')
        for directory in path:
            dir_to_create = '/'.join(current_path + [directory])
            if not os.path.exists(dir_to_create):
                os.mkdir(dir_to_create)

            current_path.append(directory)

    def get_migrations(self):
        migrations = []
        for path in os.listdir(self.migrations_dir.replace('.', '/')):
            if not path.endswith('.py'):
                continue

            module_path = f'{self.__migration_dir}.{path.replace(".py", "")}'
            module = importlib.import_module(module_path)
            migrations.append(module.Migration)

        return migrations



class JijaORMConfigTypeError(TypeError):
    APPS_MESSAGE = 'Apps must be list or tuple of jija_orm.config.App'
    CONNECTION_MESSAGE = 'Connection must be jija_orm.config.Connection, not '

    def __init__(self, *, apps=None, app=None, connection=None):
        if connection:
            super().__init__(f'Connection must be jija_orm.config.Connection, not {type(connection)}')
        else:
            error_type = None
            if app:
                error_type = f'app: {type(app)}'

            if apps:
                error_type = f'apps: {type(apps)}'

            super().__init__(f'{self.APPS_MESSAGE}, got {error_type}')


class AppConfigTypeError(TypeError):
    def __init__(self, models):
        super().__init__(f'Models must be list or tuple, not {type(models)}')


class AppConfigImportError(ImportError):
    def __init__(self, models_module):
        super().__init__(f'Cant import {models_module}')


class TemplateAttributeError(AttributeError):
    def __init__(self, model, field):
        super().__init__(f'{model} does not have {field} field')


class SelectorTypeError(TypeError):
    def __init__(self, template):
        super().__init__(f'Invalid type of selector argument: {template}')



from jija_orm import fields, exceptions
from jija_orm.executors import executors, comparators
from jija_orm.utils import NotSet


class Template:
    PRIORITY = NotImplemented
    NAME = NotImplemented
    COMPARABLE = False

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        self.model = None
        self.table = None

        self.params = set()

    def set_model(self, model=None, table=None):
        self.model = model
        self.table = table

    def get_table_name(self):
        table_name = self.model.get_name() if self.model else self.table
        table_name = table_name.replace('.', '"."')
        return f'"{table_name}"'

    def validate_field(self, field, value=None):
        if isinstance(value, executors.Param):
            self.params.add(value.name)
            return '{' + value.name + '}'

        validator = None
        if self.model:
            if not hasattr(self.model, field):
                raise exceptions.TemplateAttributeError(self.model, field)

            validator = getattr(self.model, field)

        if value is None:
            return 'null'

        if validator is None:
            validator = fields.get_validator(value)

        return validator.to_sql(value)

    def calculate(self):
        raise NotImplementedError()


class Select(Template):
    PRIORITY = 0
    NAME = 'select'

    def __init__(self, *args, distinct=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.distinct = distinct

    def calculate(self):
        select_items = self.parse_args() + self.parse_kwargs()
        return f'select {"distinct " if self.distinct else ""}{", ".join(select_items) if select_items else "*"} ' \
               f'from {self.get_table_name()}'

    def parse_args(self):
        template = []
        for field in self.args:
            self.validate_field(field)
            template.append(field)

        return template

    def parse_kwargs(self):
        template = []

        for name, field in self.kwargs.items():
            self.validate_field(field)
            template.append(f'{field} as {name}')

        return template


class Insert(Template):
    PRIORITY = 0
    NAME = 'insert'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse_kwargs(self):
        template = {}

        for name, value in self.kwargs.items():
            value = self.validate_field(name, value)
            template[name] = value

        return template

    def calculate(self):
        params = self.parse_kwargs()
        return f"insert into {self.get_table_name()} ({','.join(params)})" \
               f" values ({','.join(map(str, params.values()))})"


class MultipleInsert(Template):
    PRIORITY = 0
    NAME = 'insert'

    def __init__(self, rows):
        super().__init__()
        self.__rows = rows

    def get_fields(self):
        model_fields = set()
        for row in self.__rows:
            model_fields.update(row.keys())

        return list(map(lambda field: "field", model_fields))

    def parse_row(self, inserter_fields, row):
        if not isinstance(row, dict):
            raise TypeError(type(row))

        params = []
        for field_name in inserter_fields:
            value = row.get(field_name, NotSet)
            if value is NotSet:
                params.append('null')
                continue

            value = self.validate_field(field_name, value)
            params.append(str(value))

        return f'({",".join(params)})'

    def calculate(self):
        inserter_fields = self.get_fields()
        params = list(map(lambda row: self.parse_row(inserter_fields, row), self.__rows))

        return f"insert into {self.get_table_name()} ({','.join(inserter_fields)}) values {','.join(params)}"


class Update(Template):
    PRIORITY = 0
    NAME = 'update'

    def calculate(self):
        set_params = ','.join(map(lambda pair: f'"{pair[0]}" = {pair[1]}', self.kwargs.items()))
        return f'update {self.get_table_name()} set {set_params}'


class Delete(Template):
    PRIORITY = 0
    NAME = 'delete'

    def __init__(self):
        super().__init__()

    def calculate(self):
        return f'{self.NAME} from {self.get_table_name()}'


class WhereAddParam:
    AND = 'and'
    OR = 'or'


class Where(Template):
    PRIORITY = 1
    NAME = 'where'

    def __init__(self, add_param=WhereAddParam.AND, **kwargs):
        super().__init__(**kwargs)
        self.add_param = add_param

    def calculate(self):
        return f'where {f" {self.add_param} ".join(self.parse_kwargs())}'

    def parse_kwargs(self):
        template = []

        for field, value in self.kwargs.items():
            field, lookup, *_ = *field.split('__'), None

            if isinstance(value, comparators.Comparator):
                value.value = self.validate_field(field, value.value)
                query = value.calculate(field, getattr(self.model, field) if hasattr(self.model, field) else None)
            else:
                value = self.validate_field(field, value)
                query = f'"{field}" {self.parse_lookup(lookup)} {value}'

            template.append(query)
        return template

    @staticmethod
    def parse_lookup(lookup):
        if lookup is None:
            return '='
        elif lookup == 'lt':
            return '<'
        elif lookup == 'lte':
            return '<='
        elif lookup == 'gt':
            return '>'
        elif lookup == 'gte':
            return '>='
        elif lookup == 'not':
            return '!='
        elif lookup == 'startswith':
            return 'like'
        else:
            raise TypeError('invalid lookup')

    # def __iand__(self, other):
    #     pass

    # def __ior__(self, other):
    #     pass

    # def __ixor__(self, other):
    #     pass



from jija_orm import config
from jija_orm.migrator import selectors
from jija_orm.migrator import templates


class Migrator:
    @classmethod
    async def migrate(cls):
        # tables_migrations = await cls.__get_tables_migrations()
        # print(tables_migrations) # TODO

        models_migrations = cls.__get_models_migrations()
        exist_migrations = cls.__get_exist_migrations()
        migrations_by_migrations = cls.__get_differences(models_migrations, exist_migrations)
        await cls.__write_migrations(migrations_by_migrations)

    @classmethod
    async def __get_tables_migrations(cls):
        apps = config.JijaORM.get_apps_names()

        migrations = []
        for app in apps:
            migration = await templates.AppMigration.from_table(app)
            migration and migrations.append(migration)

        return migrations

    @classmethod
    def __get_models_migrations(cls):
        apps_models = config.JijaORM.get_models()

        migrations = []
        for app_name in apps_models:
            migration = templates.AppMigration.from_models(app_name, apps_models[app_name])
            migration and migrations.append(migration)

        return migrations

    @classmethod
    def __get_exist_migrations(cls):
        raw_app_migrations = config.JijaORM.get_migrations()

        migrations = []
        for app_name in raw_app_migrations:
            app_migration = None
            for raw_migration in raw_app_migrations[app_name]:
                migration = templates.AppMigration.from_migration(app_name, raw_migration)
                app_migration = templates.AppMigration.sqwash(app_migration, migration)

            app_migration and migrations.append(app_migration)

        return migrations

    @staticmethod
    def __get_differences(one, two):
        one_dict = dict(map(lambda mig: (mig.app, mig), one))
        two_dict = dict(map(lambda mig: (mig.app, mig), two))

        differences = []
        for name in {*one_dict, *two_dict}:
            difference = templates.AppMigration.difference(one_dict.get(name), two_dict.get(name))
            if difference:
                differences.append(difference)

        return differences

    @staticmethod
    async def __write_migrations(migrations: list[templates.AppMigration]):
        for migration in migrations:
            await migration.write()

    @classmethod
    async def __get_migrations_to_execute(cls) -> list[templates.Migration]:
        executed_id = (await selectors.migration_id_selector.execute())[0]['id']
        exist_migrations = config.JijaORM.get_migrations()

        migrations = []
        for app_name in exist_migrations:
            for app_migration in exist_migrations[app_name]:
                if executed_id and app_migration.id <= executed_id:
                    continue

                migrations.append(app_migration)

        return migrations

    @classmethod
    async def update(cls):
        migrations = await cls.__get_migrations_to_execute()

        for migration in migrations:
            await migration.execute()
            await selectors.id_setter.execute(id=migration.id)



import asyncio

from jija_orm import config, fields


class Param:
    def __init__(self, name):
        self.name = name


class __BaseExecutor:
    def __init__(self, model, *args, use_model=True, serializer=None):
        self.__model = model
        self.__use_model = use_model
        self.__serializer = model if use_model else serializer

        self.__params = set()
        self.templates = {}

        self.__parse_templates(args)
        self.__query = self.__generate_query()

    def __parse_templates(self, args):
        templates_list = sorted(args, key=lambda arg: arg.PRIORITY)

        for template in templates_list:
            if self.__use_model:
                template.set_model(model=self.__model)
            else:
                template.set_model(table=self.__model)

            self.templates[template.NAME] = template.calculate()
            self.__params.update(template.params)

    def __generate_query(self):
        return "\n".join(self.templates.values())

    @property
    def query(self):
        return self.__query

    def sync_execute(self, **kwargs):
        result = asyncio.run(self.execute(**kwargs))
        return result

    async def execute(self, **kwargs):
        if self.__params != set(kwargs.keys()):
            raise ValueError()

        connection = await self.get_connection()

        for key in kwargs.keys():
            kwargs[key] = fields.get_validator(kwargs[key]).to_sql(kwargs[key])

        query = self.__query.format(**kwargs)
        print(query)
        return self.__serialize(await connection.fetch(query))

    @staticmethod
    async def get_connection():
        return await config.JijaORM.get_connection()

    def __serialize(self, query_set):
        if not self.__serializer:
            return query_set

        return list(map(lambda record: self.__serializer(**record), query_set))


class Executor(__BaseExecutor):
    pass



import datetime

from jija_orm.utils import NotSet


class Field:
    PYTHON_TYPE = NotImplemented
    SQL_TYPE = NotImplemented
    EQUAL_PARAMS = ('default', 'null')

    def __init__(self, *, pk=False, default=NotSet, null=False):
        self.pk = pk
        self.null = null
        self.default = default

    # @classmethod
    # def to_python(cls, value):
    #     raise NotImplementedError()

    @classmethod
    def __validate_type(cls, value):
        if type(value) is not cls.PYTHON_TYPE:
            raise TypeError(f'Excepted {cls.PYTHON_TYPE.__name__}, got {type(value).__name__}')

    @classmethod
    def to_sql(cls, value):
        cls.__validate_type(value)
        return cls._to_sql(value)

    @classmethod
    def _to_sql(cls, value):
        raise NotImplementedError()

    def get_type(self):
        return self.SQL_TYPE

    @property
    def data(self):
        return {
            'type': self.get_type(),
            'null': self.null,
            'default': self.default,
            'pk': self.pk,
        }


class BooleanField(Field):
    PYTHON_TYPE = bool
    SQL_TYPE = 'bool'

    @classmethod
    def _to_sql(cls, value):
        return 'true' if value else 'false'


class CharField(Field):
    PYTHON_TYPE = str
    SQL_TYPE = 'varchar'
    EQUAL_PARAMS = (*Field.EQUAL_PARAMS, 'max_length')

    def __init__(self, *, max_length, **kwargs):
        super().__init__(**kwargs)
        self.max_length = max_length

    @classmethod
    def _to_sql(cls, value):
        return f"'{value}'"

    def get_type(self):
        return f'{self.SQL_TYPE}({self.max_length})'


class IntegerField(Field):
    PYTHON_TYPE = int
    SQL_TYPE = 'int4'

    @classmethod
    def _to_sql(cls, value):
        return value


class DateField(Field):
    PYTHON_TYPE = datetime.date
    SQL_TYPE = 'date'

    @classmethod
    def _to_sql(cls, value: datetime.date) -> str:
        return f"'{value.isoformat()}'"


class DatetimeField(Field):
    PYTHON_TYPE = datetime.datetime
    SQL_TYPE = 'timestamp'

    @classmethod
    def _to_sql(cls, value: datetime.datetime) -> str:
        return f"'{value.isoformat(' ')}'"


class RelatedField:
    def __init__(self, *, relation_to, on_delete, null=False):
        self.__relation_to = relation_to
        self.__on_delete = on_delete
        self.__null = null

    @property
    def relation_to(self):
        return self.__relation_to

    @property
    def on_delete(self):
        return self.__on_delete

    @property
    def null(self):
        return self.__null


class ForeignKey(RelatedField):
    pass


class OnDelete:
    CASCADE = 'cascade'
    SET_NULL = 'set null'
    SET_DEFAULT = 'set default'


def get_validator(value):
    if isinstance(value, bool):
        return BooleanField

    elif isinstance(value, str):
        return CharField

    elif isinstance(value, int):
        return IntegerField

    elif isinstance(value, datetime.date):
        return DateField

    elif isinstance(value, datetime.datetime):
        return DatetimeField

    raise TypeError(type(value))


SQL_TYPE_DICT = {
    BooleanField.SQL_TYPE: BooleanField,
    IntegerField.SQL_TYPE: IntegerField,
    CharField.SQL_TYPE: CharField,
    DateField.SQL_TYPE: DateField,
    DatetimeField.SQL_TYPE: DatetimeField
}



import inspect

from jija_orm import fields
from jija_orm.executors import query, executors, templates


class Model:
    id = fields.IntegerField(pk=True)

    manager = query.QueryManager

    def __init__(self, *, from_query=False, **kwargs):
        self.__from_query = from_query
        self.__state = {}
        for kwarg in kwargs:
            if hasattr(self, kwarg):
                setattr(self, kwarg, kwargs[kwarg])
                self.__state[kwarg] = kwargs[kwarg]
            else:
                raise NotImplementedError()

    @classmethod
    def init_managers(cls, real_class=None):
        for name, attr in cls.__dict__.items():
            if inspect.isclass(attr) and issubclass(attr, query.QueryManager):
                setattr(real_class, name, attr(real_class or cls))

        if hasattr(cls.__base__, 'init_managers'):
            cls.__base__.init_managers(cls)

    @classmethod
    def get_fields(cls):
        model_fields = {}

        for name, attr in cls.__dict__.items():
            if isinstance(attr, fields.Field):
                model_fields[name] = attr

        if hasattr(cls.__base__, 'get_fields'):
            model_fields.update(cls.__base__.get_fields())

        return model_fields

    @classmethod
    def get_constraints(cls) -> dict[str, fields.RelatedField]:
        model_fields = {}

        for name, attr in cls.__dict__.items():
            if isinstance(attr, fields.RelatedField):
                model_fields[name] = attr

        if hasattr(cls.__base__, 'get_constraints'):
            model_fields.update(cls.__base__.get_constraints())

        return model_fields

    async def save(self):
        to_update = {}
        for arg in self.__state:
            real_arg = getattr(self, arg)
            if self.__state[arg] != real_arg:
                to_update[arg] = real_arg

        if not to_update:
            return

        executor = executors.Executor(
            self.__class__,
            templates.Update(**to_update),
            templates.Where(id=self.id)
        )

        return await executor.execute()

    @classmethod
    def get_name(cls):
        return cls.__name__.lower()

    @classmethod
    def select(cls, *args, **kwargs):
        return cls.manager.select(*args, **kwargs)

    @classmethod
    def update(cls, *args, **kwargs):
        return cls.manager.update(*args, **kwargs)

    @classmethod
    def where(cls, *args, **kwargs):
        return cls.manager.where(*args, **kwargs)

    @classmethod
    def multiple_create(cls, rows):
        return cls.manager.multiple_insert(rows)

    def __await__(self):
        return self.manager.insert(self.__state)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.id}>'



import datetime
from typing import Optional

import aiofile

from jija_orm.migrator import selectors
from jija_orm import config, fields
from jija_orm import utils


class Action:
    CREATE = 'create'
    DROP = 'drop'
    ALTER = 'alter'


class __BaseMigrator:
    def __init__(self, name: str, action: str, childes: list):
        self.__name = name
        self.__action = action
        self.__childes = childes

    @property
    def name(self):
        return self.__name

    @property
    def action(self):
        return self.__action

    @property
    def childes(self):
        return self.__childes

    @property
    def childes_dict(self):
        return dict(map(lambda child: (child.name, child), self.__childes))

    def copy(self):
        return self.__class__(self.name, self.action, list(map(lambda child: child.copy(), self.childes)))

    @classmethod
    def sqwash(
            cls,
            oldest: Optional['__BaseMigrator'],
            newest: Optional['__BaseMigrator']
    ) -> Optional['__BaseMigrator']:

        if not newest or not oldest:
            if not newest and not oldest:
                return

            if newest and newest.action != Action.CREATE:
                raise AttributeError

            to_return = newest or oldest
            return cls(to_return.name, to_return.action, list(map(lambda child: child.copy(), to_return.childes)))

        if not type(newest) == type(oldest):
            raise TypeError(type(newest), type(oldest))

        if oldest.action == Action.CREATE:
            if newest.action == Action.DROP:
                return None

        if oldest.action == Action.DROP:
            if newest.action == Action.CREATE:
                return cls(newest.name, Action.CREATE, newest.childes.copy())

            if oldest.action == Action.DROP:
                return cls(newest.name, Action.DROP, [])

            raise AttributeError()

        if oldest.action == Action.ALTER:
            if newest.action == Action.CREATE:
                raise AttributeError()

            if newest.action == Action.DROP:
                return cls(newest.name, Action.DROP, [])

        newest_dict = newest.childes_dict
        oldest_dict = oldest.childes_dict

        childes = []
        for field_name in {*oldest_dict, *newest_dict}:
            child_class = newest_dict.get(field_name) or oldest_dict.get(field_name)

            field = child_class.sqwash(oldest_dict.get(field_name), newest_dict.get(field_name))
            if field:
                childes.append(field)

        return cls(newest.name, oldest.action, childes)

    @classmethod
    def difference(
            cls,
            newest: Optional['__BaseMigrator'],
            oldest: Optional['__BaseMigrator']
    ) -> Optional['__BaseMigrator']:
        if newest == oldest:
            return

        if not newest:
            if oldest.action == Action.DROP:
                return

            return cls(oldest.name, Action.DROP, [])

        if not oldest:
            if newest.action != Action.CREATE:
                raise AttributeError()

            return cls(newest.name, Action.CREATE, list(map(lambda child: child.copy(), newest.childes)))

        if not type(newest) == type(oldest):
            raise TypeError(type(newest), type(oldest))

        if newest.action == Action.CREATE:
            if oldest.action == Action.DROP:
                return cls(newest.name, Action.CREATE, list(map(lambda child: child.copy(), newest.childes)))

            if oldest.action == Action.ALTER:
                raise AttributeError()

        if newest.action == Action.DROP:
            if oldest.action == Action.DROP:
                return

            return cls(newest.name, Action.DROP, [])

        if newest.action == Action.ALTER:
            if oldest.action == Action.CREATE:
                raise AttributeError()

            if oldest.action == Action.DROP:
                return cls(newest.name, Action.ALTER, list(map(lambda child: child.copy(), newest.childes)))

        newest_dict = newest.childes_dict
        oldest_dict = oldest.childes_dict

        childes = []
        for field_name in {*oldest_dict, *newest_dict}:
            child_class = newest_dict.get(field_name) or oldest_dict.get(field_name)

            field = child_class.difference(newest_dict.get(field_name), oldest_dict.get(field_name))
            if field:
                childes.append(field)

        if not childes:
            return

        if oldest.action == Action.DROP:
            action = Action.CREATE
        else:
            action = Action.ALTER

        return cls(newest.name, action, childes)

    def get_constructor(self):
        childes = []
        for child in self.childes:
            childes.extend(map(lambda line: f'        {line}', child.get_constructor()))

        return [
            f'templates.{self.__class__.__name__}(',
            f'    name="{self.name}",',
            f'    action=templates.Action.{self.action.upper()},',
            f'    childes=[',
            *childes,
            f'    ]',
            '),'
        ]

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
               self.name == other.name and \
               self.action == other.action and \
               self.__eq_childes(other)

    def __eq_childes(self, other):
        self_childes = self.childes_dict
        other_childes = other.childes_dict

        for child in self_childes:
            other_child = other_childes.pop(child, None)
            if self_childes[child] != other_child:
                return False

        if other_childes:
            return False

        return True

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.__action} {self.__name}>'

    def get_query(self, *args, **kwargs):
        raise NotImplementedError()


class AppMigration:
    def __init__(self, app, migrations):
        self.__app = app
        self.__migrations = migrations

    @property
    def app(self):
        return self.__app

    @property
    def migrations(self):
        return self.__migrations

    @property
    def migrations_dict(self):
        return dict(map(lambda child: (child.name, child), self.__migrations))

    @classmethod
    async def from_table(cls, app: str) -> Optional['AppMigration']:
        migrations = []

        for table_record in await selectors.table_selector.execute(app=app):
            migrations.append(await ModelMigration.from_table(table_record['name']))

        if not migrations:
            return

        return cls(app, migrations)

    @classmethod
    def from_models(cls, app: str, models: list) -> Optional['AppMigration']:
        migrations = []

        for model in models:
            migrations.append(ModelMigration.from_model(model))

        if not migrations:
            return

        return cls(app, migrations)

    @classmethod
    def from_migration(cls, app: str, migration: 'Migration'):
        migrations = []

        for command in migration.commands:
            migrations.append(command)

        return cls(app, migrations)

    @classmethod
    def sqwash(cls, oldest: Optional['AppMigration'], newest: Optional['AppMigration']) -> Optional['AppMigration']:
        if not oldest or not newest:
            return oldest or newest

        newest_dict = newest.migrations_dict
        oldest_dict = oldest.migrations_dict

        migrations = []
        for migration_name in {*oldest_dict, *newest_dict}:
            field = ModelMigration.sqwash(oldest_dict.get(migration_name), newest_dict.get(migration_name))
            field and migrations.append(field)

        return cls(newest.app, migrations) if migrations else None

    @classmethod
    def difference(cls, newest: Optional['AppMigration'], oldest: Optional['AppMigration']) -> Optional['AppMigration']:
        if newest == oldest:
            return

        if not newest:
            return

        if not oldest:
            return cls(newest.app, list(map(lambda migration: migration.copy(), newest.migrations)))

        newest_dict = newest.migrations_dict
        oldest_dict = oldest.migrations_dict

        differences = []
        for name in {*newest_dict, *oldest_dict}:
            difference = ModelMigration.difference(newest_dict.get(name), oldest_dict.get(name))
            difference and differences.append(difference)

        return cls(newest.app, differences) if differences else None

    async def write(self):
        migration_dir = config.JijaORM.get_migrations_dir(self.app)
        migration_id, constructor = Migration.get_constructor(self.migrations)

        async with aiofile.async_open(migration_dir.replace('.', '/') + f'/{migration_id}.py', 'w') as file:
            await file.write(constructor)

    def __repr__(self):
        return f'<AppMigration {self.app}>'


class ModelMigration(__BaseMigrator):
    @classmethod
    def from_model(cls, model):
        return cls(
            model.get_name(), Action.CREATE,
            [*cls.__get_fields(model), *cls.__get_constraints(model)]
        )

    @classmethod
    def __get_fields(cls, model):
        raw_fields = model.get_fields()
        return list(map(lambda field_name: FieldMigration.from_field(field_name, raw_fields[field_name]), raw_fields))

    @classmethod
    def __get_constraints(cls, model):
        raw_constraints = model.get_constraints()

        childes = []
        for constraint_name in raw_constraints:
            childes.append(FieldMigration.from_field(f'{constraint_name}_id', raw_constraints[constraint_name]))
            childes.append(ConstraintMigration.from_field(constraint_name, raw_constraints[constraint_name]))

        return childes
        # return list(map(lambda field_name: FieldMigration.from_field(field_name, raw_fields[field_name]), raw_fields))

    @classmethod
    async def from_table(cls, table):
        raw_columns = await cls.__get_field_data(table)

        return cls(
            table, Action.CREATE,
            list(map(lambda column: FieldMigration.from_column(column), raw_columns))
        )

    @staticmethod
    async def __get_field_data(table):
        return await (await config.JijaORM.get_connection()).fetch(f"""
            select
                column_name as "name",
                column_default as "default",
                is_nullable as "null",
            
                (case when udt_name = 'varchar' then concat(udt_name, '(', character_maximum_length, ')')
                 else udt_name end) as type,
            
                column_name = (select pg_attribute.attname::text
                 from pg_index,
                      pg_class,
                      pg_attribute,
                      pg_namespace
                 where pg_class.oid = '{table}'::regclass
                   and indrelid = pg_class.oid
                   AND nspname = 'public'
                   AND pg_class.relnamespace = pg_namespace.oid
                   AND pg_attribute.attrelid = pg_class.oid
                   AND pg_attribute.attnum = any (pg_index.indkey)
                   AND indisprimary) as pk
            
            from information_schema.columns
            where table_name = '{table}'
                """)

    async def execute(self):
        query = self.get_query()

        print(query)
        connection = await config.JijaORM.get_connection()
        await connection.execute(query)

    def get_query(self):
        if self.action == Action.DROP:
            return f'drop table "{self.name}"'

        childes = self.__sort_childes()
        childes_query = ', '.join(map(lambda field: field.get_query(self.action), childes))

        if self.action == Action.CREATE:
            return f'create table "{self.name}" ({childes_query})'

        else:
            return f'alter table "{self.name}" {childes_query}'

    def __sort_childes(self) -> list:
        childes = sorted(self.childes, key=lambda _child: isinstance(_child, FieldMigration))

        if self.action == Action.CREATE:
            return childes

        constraint_indexes = {}
        field_indexes = {}
        for index, child in enumerate(childes):
            if isinstance(child, FieldMigration):
                field_indexes[child.name.replace('_id', '')] = index
            else:
                constraint_indexes[child.name] = index

        if not constraint_indexes:
            return childes

        for name in constraint_indexes:
            pair = [childes[field_indexes[name]], childes[constraint_indexes[name]]]
            if childes[constraint_indexes[name]].action == Action.DROP:
                pair = list(reversed(pair))

            for attr in pair:
                childes.remove(attr)

            childes.extend(pair)

        return childes


class FieldMigration(__BaseMigrator):
    @classmethod
    def from_field(cls, name, field):
        if isinstance(field, fields.Field):
            childes = [
                    AttributeMigration('pk', field.pk),
                    AttributeMigration('null', field.null),
                    AttributeMigration('type', field.get_type()),
                ]

            if field.default is not utils.NotSet:
                childes.append(AttributeMigration('default', field.default))

            return cls(name, Action.CREATE, childes)

        if isinstance(field, fields.RelatedField):
            return cls(
                name,
                Action.CREATE,
                [
                    AttributeMigration('null', field.null),
                    AttributeMigration('type', fields.IntegerField.SQL_TYPE),
                ]
            )

        raise TypeError(type(field))

    @classmethod
    def from_column(cls, column):
        return cls(
            column['name'],
            Action.CREATE,
            [
                AttributeMigration('pk', column['pk']),
                AttributeMigration('default', column['default']),
                AttributeMigration('null', column['null']),
                AttributeMigration('type', column['type']),
            ]
        )

    # def get_model_create_query(self):
    #     attrs = sorted(self.childes, key=lambda attr: attr.priority, reverse=True)
    #     attrs_query = ' '.join(list(filter(lambda attr: attr, map(lambda attr: attr.get_create_query(), attrs))))
    #     return f"{self.name} {attrs_query}"

    def get_query(self, model_action):
        if self.action == Action.DROP:
            return f'drop column "{self.name}"'

        elif self.action == Action.ALTER:
            return self.__get_alter_query()

        else:
            return self.__get_create_query(model_action)

    def __get_alter_query(self):
        query_lines = []
        for attr in self.childes:
            child_query = attr.get_query(Action.ALTER)
            child_query and query_lines.append(f'alter column "{self.name}" {child_query}')

        return ', '.join(query_lines)

    def __get_create_query(self, model_action):
        attrs = sorted(self.childes, key=lambda attr: attr.priority)

        attrs_query = ' '.join(
            filter(lambda attr: attr,
                   map(lambda attr: attr.get_query(self.action),
                       attrs)
                   )
        )
        query = f'"{self.name}" {attrs_query}'

        if model_action != Action.CREATE:
            query = f'add column {query}'

        return query


class ConstraintMigration(__BaseMigrator):
    @classmethod
    def from_field(cls, name, field):
        return cls(
            name, Action.CREATE,
            [
                ConstraintAttributeMigration('field', f'{name}_id'),
                ConstraintAttributeMigration('relation_to', field.relation_to),
                ConstraintAttributeMigration('on_delete', field.on_delete),
            ]
        )

    @classmethod
    def from_column(cls):
        raise NotImplementedError()

    def get_query(self, model_action):
        attrs = sorted(self.childes, key=lambda attr: attr.priority)
        attrs_query = " ".join(map(lambda attr: attr.get_query(self.action), attrs))

        query = f'constraint "{self.name}" {attrs_query}'

        if self.action == Action.CREATE:
            action_query = 'add'
        else:
            action_query = self.action

        if model_action != Action.CREATE:
            query = f'{action_query} {query}'

        return query


class __BaseAttributeMigration:
    PRIORITY: dict = NotImplemented

    def __init__(self, name, value):
        self.__name = name
        self.__value = value
        self.__priority = self.PRIORITY.get(name, 999)

    @property
    def name(self):
        return self.__name

    @property
    def value(self):
        return self.__value

    @property
    def priority(self):
        return self.__priority

    def copy(self):
        return self.__class__(self.name, self.value)

    @classmethod
    def sqwash(cls, oldest: Optional['__BaseAttributeMigration'], newest: Optional['__BaseAttributeMigration']
               ) -> '__BaseAttributeMigration':

        return newest or oldest

    @classmethod
    def difference(cls, newest: Optional['__BaseAttributeMigration'], oldest: Optional['__BaseAttributeMigration']
                   ) -> Optional['__BaseAttributeMigration']:

        if not newest:
            return

        if not oldest:
            return newest

        if not type(newest) == type(oldest):
            raise TypeError(type(newest), type(oldest))

        if newest.value == oldest.value:
            return

        return cls(newest.name, newest.value)

    def get_query(self, model_action):
        raise NotImplementedError()

    def get_constructor(self):
        return [f'templates.{self.__class__.__name__}('f'name="{self.name}", value={self.__get_constructor_value()}),']

    def __get_constructor_value(self):
        if self.value is None or isinstance(self.value, bool):
            return self.value

        if isinstance(self.value, str):
            return f'"{self.value}"'

        return self.value

    def __eq__(self, other: '__BaseAttributeMigration') -> bool:
        return isinstance(other, self.__class__) and \
               self.name == other.name and \
               self.value == other.value and \
               self.priority == other.priority

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.name} {self.value}>'


class AttributeMigration(__BaseAttributeMigration):
    PRIORITY = {'type': 0}
    ATTRS_TRANSLATOR = {
        'type': 'type {}',
        'pk': '{} primary key',
        'null': '{} not null',
        'default': '{} default'
    }

    def get_query(self, field_action):
        return getattr(self, f'_{self.__class__.__name__}__get_{self.name}')(field_action)

    def __get_type(self, field_action):
        if field_action == Action.CREATE:
            return self.value

        return f'type {self.value}'

    def __get_null(self, field_action):
        if field_action == Action.ALTER:
            return f'{"drop" if self.value else "set"} not null'

        if not self.value:
            return 'not null'

    def __get_pk(self, field_action):
        if field_action == Action.ALTER:
            return f'{"set" if self.value else "drop"} primary key'

        if self.value:
            return 'primary key generated by default as identity'

    def __get_default(self, field_action):
        if self.value is None:
            query = 'default null'
        elif isinstance(self.value, int):
            query = f'default {self.value}'
        else:
            query = f'default "{self.value}"'

        if field_action == Action.CREATE:
            return query
        return f'set {query}'


class ConstraintAttributeMigration(__BaseAttributeMigration):
    PRIORITY = {
        'field': 0,
        'relation_to': 1,
        'on_delete': 2
    }

    ATTRS_TRANSLATOR = {
        'field': 'foreign key ({})',
        'relation_to': 'references {}(id)',
        'on_delete': 'on delete {}'
    }

    def __init__(self, name, value):
        if name == 'relation_to':
            value = value.get_name()

        super().__init__(name, value)

    def get_query(self, model_action):
        return self.ATTRS_TRANSLATOR[self.name].format(self.value)


class Migration:
    id: int = NotImplemented
    commands: tuple[ModelMigration] = NotImplemented

    @classmethod
    async def execute(cls):
        for command in cls.commands:
            await command.execute()

    @staticmethod
    def get_constructor(commands):
        commands_lines = []
        for command in commands:
            commands_lines.extend(map(lambda line: f'        {line}', command.get_constructor()))

        now = datetime.datetime.now()
        migration_id = int(datetime.datetime.now().timestamp() * 1000000)
        template = [
            f'# Created by jija-orm at {now.replace(microsecond=0)}',
            'from jija_orm.migrator import templates',
            '',
            '',
            f'class Migration(templates.Migration):',
            f'    id = {migration_id}',
            '    commands = (',
            *commands_lines,
            '    )',
            ''
        ]

        return migration_id, '\n'.join(template)



import asyncio
import asyncpg


class Connection:
    def __init__(self, host, port, user, password, database):
        self.__host = host
        self.__port = port
        self.__user = user
        self.__password = password
        self.__database = database

    def check(self):
        asyncio.run(self.async_check())

    async def async_check(self):
        await asyncpg.connect(f'postgresql://{self.__user}:{self.__password}@{self.__host}/{self.__database}')

    async def get_connection(self) -> asyncpg.connection.Connection:
        return await asyncpg.connect(f'postgresql://{self.__user}:{self.__password}@{self.__host}/{self.__database}')



import os
import pathlib

import asyncpg

from jija_orm import config
from jija_orm import exceptions


class JijaORM:
    APPS = None
    CONNECTION = None
    PROJECT_DIR = None

    def __init__(self, *, apps=None, connection=None, project_dir=None):
        self.__set_params(apps=apps, connection=connection, project_dir=project_dir)

        for app in JijaORM.APPS:
            app.check(JijaORM.PROJECT_DIR)
            app.load()

        JijaORM.CONNECTION.check()

    @classmethod
    async def async_init(cls, *, apps=None, connection=None, project_dir=None):
        cls.__set_params(apps=apps, connection=connection, project_dir=project_dir)

        for app in JijaORM.APPS:
            app.check(JijaORM.PROJECT_DIR)
            await app.async_load()

        await JijaORM.CONNECTION.async_check()

    @classmethod
    def __set_params(cls, *, apps, connection, project_dir):
        JijaORM.APPS = cls.__validate_apps(apps)
        JijaORM.CONNECTION = cls.__validate_connection(connection)
        JijaORM.PROJECT_DIR = project_dir or pathlib.Path(os.getcwd())

    @staticmethod
    def __validate_apps(apps):
        if not apps:
            return []

        if not isinstance(apps, (list, tuple)):
            raise exceptions.JijaORMConfigTypeError(apps=apps)

        for app in apps:
            if not isinstance(app, config.App):
                raise exceptions.JijaORMConfigTypeError(app=app)

        return apps

    @staticmethod
    def __validate_connection(connection):
        if not isinstance(connection, config.Connection):
            raise exceptions.JijaORMConfigTypeError(connection=connection)

        return connection

    @classmethod
    async def get_connection(cls) -> asyncpg.connection.Connection:
        return await cls.CONNECTION.get_connection()

    @classmethod
    def get_models(cls) -> dict[str, list]:
        models = {}

        for app in cls.APPS:
            models[app.name] = list(app.models.values())

        return models

    @classmethod
    def get_migrations_dir(cls, app_name):
        for app in cls.APPS:
            if app.name == app_name:
                return app.migrations_dir

        raise AttributeError(app)

    @classmethod
    def get_migrations(cls) -> dict[str, list]:
        migrations = {}

        for app in cls.APPS:
            migrations[app.name] = app.get_migrations()

        return migrations

    @classmethod
    def get_apps_names(cls) -> list[str]:
        return list(map(lambda app: app.name, cls.APPS))



from jija_orm.executors import templates
from jija_orm.executors import executors


class QueryManager:
    def __init__(self, model_class, class_templates=None):
        self.__model_class = model_class
        self.__templates = self.fix_templates(class_templates)

    @staticmethod
    def fix_templates(class_templates):
        if not class_templates:
            return []

        templates_dict = {}
        for template in class_templates:
            if templates_dict.get(template.PRIORITY):
                if template.COMPARABLE is False:
                    templates_dict[template.PRIORITY] = template
                else:
                    raise NotImplementedError()

            else:
                templates_dict[template.PRIORITY] = template

        return list(templates_dict.values())

    def select(self, *args, **kwargs):
        class_templates = self.__templates.copy()
        class_templates.append(templates.Select(*args, **kwargs))
        return self.__class__(self.__model_class, class_templates)

    def update(self, *args, **kwargs):
        class_templates = self.__templates.copy()
        class_templates.append(templates.Update(*args, **kwargs))
        return self.__class__(self.__model_class, class_templates)

    def insert(self, **kwargs):
        class_templates = self.__templates.copy()
        class_templates.append(templates.Insert(**kwargs))
        return self.__class__(self.__model_class, class_templates)

    def multiple_insert(self, rows):
        # TODO need freeze?
        class_templates = self.__templates.copy()
        class_templates.append(templates.MultipleInsert(rows))
        return self.__class__(self.__model_class, class_templates)

    def where(self, *args, **kwargs):
        if self.__templates:
            class_templates = self.__templates.copy()
        else:
            class_templates = [templates.Select()]

        class_templates.append(templates.Where(*args, **kwargs))
        return self.__class__(self.__model_class, class_templates)

    def __await__(self):
        async def wrapper():
            executor = executors.Executor(self.__model_class, *self.__templates)
            return await executor.execute()

        return wrapper().__await__()



from jija_orm import models


class ModelSerializer:
    class Empty:
        pass

    def __init__(self, model: 'models.Model'):
        self.__model = model
        self.__fields = self.get_fields()

    def get_fields(self):
        return self.__model.get_fields()

    def serialize(self, record):
        data = {}
        for field in self.__fields:
            attr = record.get(field, self.Empty)
            if attr is self.Empty:
                continue

            data[field] = attr

        return models.SerializedModel(**data)



from setuptools import setup

setup(
    name='jija-orm',
    version='0.0.3-2',
    description='',
    packages=[
        'jija_orm',
        'jija_orm.config',
        'jija_orm.migrator',
        'jija_orm.executors',
    ],
    author='Kain',
    author_email='kainedezz.2000@gmail.com',
    zip_safe=False,

    install_requires=[
        'asyncpg==0.25.0',
        'aiofile==3.8.1'
    ]
)
